indexing

	description: "Class that combines two TOKEN_TABLE objects into a probabily file. It uses Paul Graham's calculation. Note that the classifier uses a different algorithm, so it uses the token counts, but not the calculated probability."


class

	PG_PROBABILITY_TABLE


inherit

	PROBABILITY_TABLE


creation

	make_combine


feature {NONE} -- Initialization

	make_combine (a_spam_table, a_nonspam_table: TOKEN_TABLE) is
			-- Combine a spam and nonspam table into a single hash with
			-- for each token the probability that it indicates spam.
			-- See http://www.paulgraham.com/spam.html for more detail.
		require
			spam_table_not_void: a_spam_table /= Void
			nonspam_table_not_void: a_nonspam_table /= Void
		local
			c: DS_HASH_TABLE_CURSOR [INTEGER, STRING]
			occurrences_spam,
			occurrences_nonspam: INTEGER
			spam_tokens,
			nonspam_tokens: DS_HASH_TABLE [INTEGER, STRING]
			p: DOUBLE
			total_tokens,
			seen_tokens: INTEGER
			percentage_done,
			prev_percentage_done: INTEGER
			data_point: TOKEN_PROBABILITY
		do
			spam_messages_count := a_spam_table.message_count
			nonspam_messages_count := a_nonspam_table.message_count
			create tokens.make ((a_spam_table.tokens.count + a_nonspam_table.tokens.count) * 2)

			-- First get all spam tokens
			stdout.put_string ("scanning spam tokens...%N")
			from
				total_tokens := a_spam_table.tokens.count
				seen_tokens := 0
				prev_percentage_done := -1
				nonspam_tokens := a_nonspam_table.tokens
				c := a_spam_table.tokens.new_cursor
				c.start
			until
				c.after
			loop
				percentage_done := (seen_tokens * 100) // total_tokens
				if percentage_done /= prev_percentage_done then
					stdout.put_string ("Completed ")
					stdout.put_string (percentage_done.out)
					stdout.put_string ("%%%R")
					stdout.flush
					prev_percentage_done := percentage_done
				end
				-- calculate `b'
				occurrences_spam := c.item
				-- calculate `g'
				nonspam_tokens.search (c.key)
				if nonspam_tokens.found then
					-- This is what Paul Graham does:
					-- nonspam token is more important, so don't miss it
					-- when it occurs at least three times.
					--occurrences_nonspam := 2 * nonspam_tokens.found_item
					-- Using Gary Robinson's formula, we can handle low
					-- data points better, so we just say:
					occurrences_nonspam := nonspam_tokens.found_item
				else
					occurrences_nonspam := 0
				end
				-- Skip tokens that don't occur often.
				-- These are for example message identifiers and such.
				-- Paul Graham set's `minimum_occurrences' to 5, we use 0,
				-- we don't want to miss any data points.
				if occurrences_spam + occurrences_nonspam >= minimum_occurrences then
					p := spam_probability (occurrences_spam, occurrences_nonspam, a_spam_table.message_count, a_nonspam_table.message_count)
					create data_point.make (c.key, p, occurrences_spam, occurrences_nonspam)
					tokens.put (data_point, c.key)
					tokens.search (c.key)
					-- Can this happen?
-- 					if not Result.found_item.token.is_equal (c.key) then
-- 						std.error.put_string ("Wow!%N")
-- 					end
				end
				c.forth
				seen_tokens := seen_tokens + 1
			end
			stdout.put_string ("Completed 100%%%N")

			-- Next the nonspam tokens we haven't yet processed
			stdout.put_string ("scanning nonspam tokens...%N")
			from
				total_tokens := a_nonspam_table.tokens.count
				seen_tokens := 0
				prev_percentage_done := -1
				spam_tokens := a_spam_table.tokens
				c := a_nonspam_table.tokens.new_cursor
				c.start
			until
				c.after
			loop
				percentage_done := (seen_tokens * 100) // total_tokens
				if percentage_done /= prev_percentage_done then
					stdout.put_string ("Completed ")
					stdout.put_string (percentage_done.out)
					stdout.put_string ("%%%R")
					stdout.flush
					prev_percentage_done := percentage_done
				end
				-- calculate `g'
				occurrences_nonspam := c.item
				-- calculate `b'
				spam_tokens.search (c.key)
				if not spam_tokens.found then
					-- Skip nonspam tokens that occur less than 3 times in
					-- Paul Graham's original formula. We don't need that,
					-- because we our spaminess formula handles low data
					-- points better.
					if 2*occurrences_nonspam >= minimum_occurrences then
						-- replace by 0.01 to speed up the loop??
						p := spam_probability (0, occurrences_nonspam, a_spam_table.message_count, a_nonspam_table.message_count)
						create data_point.make (c.key, p, occurrences_spam, occurrences_nonspam)
						tokens.put (data_point, c.key)
					end
				end
				c.forth
				seen_tokens := seen_tokens + 1
			end
			stdout.put_string ("Completed 100%%%N")

			stdout.put_string ("Calculating no data points value...%N")
			calculate_no_data_points_value
		end


feature -- Access

	minimum_occurrences: INTEGER is 0
			-- Minimum occurrences of a data point to end up on our
			-- probability file. Paul Graham sets it to 5, but we're
			-- using a different formula, and don't need a cut off.

	min_data_points: INTEGER is 10
			-- How many data points a token must have to be taken into
			-- account by `calculate_no_data_points_value'.


feature -- Formula's

	calculate_no_data_points_value is
			-- Calculate the value of `no_data_points_value'. It is the
			-- number to be used for tokens that do not occur in
			-- `tokens'.
			-- It is calculated by taking the averate of the probability
			-- of all tokens in `tokens' that have more than
			-- `min_data_points' data points.
		local
			data_point: TOKEN_PROBABILITY
			number_of_points: INTEGER
			sum: DOUBLE
		do
			from
				tokens.start
			until
				tokens.after
			loop
				data_point := tokens.item_for_iteration
				if data_point.bad_count + data_point.good_count > min_data_points then
					sum := sum + data_point.probability
					number_of_points := number_of_points + 1
				end
				tokens.forth
			end
			if number_of_points = 0 then
				no_data_points_value := 0.5
			else
				no_data_points_value := sum / number_of_points
			end
		end

	spam_probability (
				spam_occurrences,
				nonspam_occurrences: INTEGER;
				spam_message_count,
				nonspam_message_count: INTEGER): DOUBLE is
			-- Paul Graham's formula for calculating the probability that
			-- a token is spam. The nearer the number is to 1, the more
			-- likely it is to indicate spam.
			-- The formula is modified according to his comments in
			-- http://www.paulgraham.com/better.html.
		require
			spam_occurrences_not_negative: spam_occurrences >= 0
			nonspam_occurrences_not_negative: nonspam_occurrences >= 0
			spam_message_count_positive: spam_message_count > 0
			nonspam_message_count_positive: nonspam_message_count > 0
		local
			spam_chance: DOUBLE
			nonspam_chance: DOUBLE
		do
			if spam_occurrences = 0 then
				if nonspam_occurrences > 10 then
					Result := 0.009
				else
					Result := 0.01
				end
			elseif nonspam_occurrences = 0 then
				if spam_occurrences > 10 then
					Result := 0.9999
				else
					Result := 0.9998
				end
			else
				spam_chance := spam_occurrences / spam_message_count
				if spam_chance > 1 then
					spam_chance := 1
				end
				nonspam_chance := nonspam_occurrences / nonspam_message_count
				if nonspam_chance > 1 then
					nonspam_chance := 1
				end
				Result := spam_chance / (spam_chance + nonspam_chance)
			end
		ensure
			valid_probability: Result >= 0 and Result <= 1
		end


end
