indexing

	description: "Class that scans email and builds a hash table for the tokens."


class

	SCAN


inherit

	EMC_BASE

	KL_SHARED_ARGUMENTS
		export
			{NONE} all
		end

	TOKENIZE_ROUTINES

	ST_FORMATTING_ROUTINES
		export
			{NONE} all
		end


creation

	make


feature {NONE} -- Initialization

	make is
		local
			spam_table,
			nonspam_table: TOKEN_TABLE
			probability_table: PG_PROBABILITY_TABLE
		do
			stdout.put_string ("mailbox scanner " + version + "  " + copyright + ".%N")
			if Arguments.argument_count = 1 and then Arguments.argument (1).is_equal ("-h") then
				print_help
			else
				read_configuration
				check_configuration
				check_mailboxes (configuration.spam_mailboxes)
				check_mailboxes (configuration.nonspam_mailboxes)
				stdout.put_string ("Scanning ")
				stdout.put_string (configuration.spam_mailboxes.count.out)
				stdout.put_string (" mailboxes for spam tokens...%N")
				spam_table := tokenize_mailboxes (configuration.spam_mailboxes)
				--dump_tokens ("spam.txt", spam_table.message_count, spam_table.tokens)
				stdout.put_string ("Scanning ")
				stdout.put_string (configuration.nonspam_mailboxes.count.out)
				stdout.put_string (" mailboxes for nonspam tokens...%N")
				nonspam_table := tokenize_mailboxes (configuration.nonspam_mailboxes)
				stdout.put_string ("Create probability table...%N")
				create probability_table.make_combine (spam_table, nonspam_table)
				stdout.put_string ("Dumping probability file...%N")
				probability_table.save ("probabilities.txt", configuration.minimum_occurrences)
				print_statistics (probability_table)
				help_if_file_must_be_moved
				stdout.put_string ("Finished.%N")
			end
		end


feature {NONE} -- Help

	help_if_file_must_be_moved is
			-- Print a message if the probabilities.txt file must be
			-- moved to the directory where emc.config was found.
		local
			path: STDC_PATH
			config_dir: STRING
		do
			create path.make_from_string (config_file_name)
			path.parse (Void)
			config_dir := path.directory
			if not current_directory.is_equal (config_dir) then
				stdout.put_string (format ("%NThe new probabilities.txt file is put in the current directory.%NIt must be moved to the directory that contains your emc.config file.%NThat directory is $s/.%N", <<config_dir>>))
			end
		end

	print_help is
		do
			stdout.put_string ("Sorry, no help yet.%N")
		end


feature -- Configuration

	check_mailboxes (a_mailboxes: DS_LIST [MAILBOX]) is
			-- Check if the passed mailboxes exist, before actually
			-- scanning them.
		require
			a_mailboxes_not_void: a_mailboxes /= Void
		local
			client: EPX_IMAP4_CLIENT
			c: DS_LINEAR_CURSOR [MAILBOX]
		do
			client := connect_to_mail_server
			c := a_mailboxes.new_cursor
			from
				c.start
			until
				c.after
			loop
				client.examine (c.item.name)
				if not client.response.is_ok then
					stderr.put_string ("Fatal error: cannot read mailbox: ")
					stderr.put_string (c.item.name)
					stderr.put_string (".%N")
					exit_with_failure
				end
				c.forth
			end
			if client.state.is_selected then
				client.close_mailbox
			end
			client.logout
			client.close
		end


feature -- Scanning

	tokenize_mailboxes (a_mailboxes: DS_LIST [MAILBOX]): TOKEN_TABLE is
			-- Scan all mailboxes. Returns a table of tokens.
		require
			a_mailboxes_not_void: a_mailboxes /= Void
		local
			c: DS_LINEAR_CURSOR [MAILBOX]
			i: INTEGER
			body_tokenizer: TOKENIZER
			client: EPX_IMAP4_CLIENT
		do
			client := connect_to_mail_server
			create body_tokenizer.make
			c := a_mailboxes.new_cursor
			from
				c.start
				i := 1
			until
				c.after
			loop
				stdout.put_string ("Scanning ")
				stdout.put_string (c.item.name)
				stdout.put_string (" (")
				stdout.put_string (i.out)
				stdout.put_string (" of ")
				stdout.put_string (a_mailboxes.count.out)
				stdout.put_string (")...%N")
				scan (client, body_tokenizer, c.item)
				i := i + 1
				c.forth
			end
			client.logout
			client.close
			create Result.make (body_tokenizer.count, body_tokenizer.hash_table)
		ensure
			token_table_returned: Result /= Void
		end

	scan (client: EPX_IMAP4_CLIENT; a_tokenizer: TOKENIZER; a_mailbox: MAILBOX) is
			-- Tokenize every message in mailbox `a_mailbox_name'.
		require
			client_not_void: client /= Void
			client_logged_in: client.state.is_authenticated or else client.state.is_selected
			tokenizer_not_void: a_tokenizer /= Void
			mailbox_not_void: a_mailbox /= Void
		local
			percentage_done,
			prev_percentage_done: INTEGER
			total_messages_to_scan: INTEGER
			parser: EPX_MIME_PARSER
			message_number: INTEGER
			highest_sequence,
			lowest_sequence: INTEGER
			header_parsed,
			body_parsed: BOOLEAN
		do
			client.examine (a_mailbox.name)
			if not client.response.is_ok then
				stderr.put_string ("Fatal error: cannot read mailbox: ")
				stderr.put_string (a_mailbox.name)
				stderr.put_string (".%N")
				exit_with_failure
			end
			assert ("Read-only mailbox.", not client.response.current_mailbox.is_writable)

			-- Determine lowest and highest sequence of messages to scan.
			lowest_sequence := a_mailbox.lowest_sequence
			highest_sequence := client.response.current_mailbox.count
			if
				a_mailbox.highest_sequence /= 0 and then
				a_mailbox.highest_sequence < highest_sequence
			then
				highest_sequence := a_mailbox.highest_sequence
			end

			-- Correct for `a_mailbox'.`max_messages' and
			-- `a_mailbox'.`percentage'.
			total_messages_to_scan := highest_sequence - lowest_sequence + 1
			if a_mailbox.max_messages /= 0 then
				if total_messages_to_scan > a_mailbox.max_messages then
					lowest_sequence := (highest_sequence - a_mailbox.max_messages) + 1
				end
			end
			total_messages_to_scan := highest_sequence - lowest_sequence + 1
			if a_mailbox.percentage /= 0 then
				if (total_messages_to_scan * 100) // client.response.current_mailbox.count > a_mailbox.percentage then
					lowest_sequence := highest_sequence - (total_messages_to_scan * ((100 - a_mailbox.percentage) / 100)).rounded
				end
			end

			total_messages_to_scan := highest_sequence - lowest_sequence + 1
			stdout.put_string ("Number of messages: ")
			stdout.put_string (client.response.current_mailbox.count.out)
			if total_messages_to_scan < client.response.current_mailbox.count then
				stdout.put_string (" (learning only ")
				stdout.put_string (total_messages_to_scan.out)
				stdout.put_string (")%N")
			end
			stdout.put_string ("%N")
			create parser.make
			from
				prev_percentage_done := -1
				message_number := highest_sequence
			invariant
				lowest_sequence >= 1
			variant
				message_number
			until
				message_number < lowest_sequence or else
				message_number > client.response.current_mailbox.count
			loop
				percentage_done := ((client.response.current_mailbox.count - message_number) * 100) // total_messages_to_scan
				if percentage_done /= prev_percentage_done then
					fd_stdout.put_string ("Completed ")
					fd_stdout.put_string (percentage_done.out)
					fd_stdout.put_string ("%%%R")
					prev_percentage_done := percentage_done
				end
				client.fetch_header (message_number)
				if client.response.is_ok then
					client.fetch_size (message_number)
					assert ("Fetch message number ok", client.response.is_ok)
					if client.response.current_message.size < a_mailbox.max_message_size then
						client.fetch_body (message_number)
						assert ("Fetch body ok", client.response.is_ok)
						if not client.response.current_message.body.is_empty then
							debug ("emc")
								--print ("Message number: ")
								--print (message_number.out)
								--print ("%N")
							end
							parser.set_input_buffer (parser.new_string_buffer (client.response.current_message.header))
							parser.parse_header
							header_parsed :=
								not parser.syntax_error and then
								parser.parsing_errors = 0
							debug ("emc")
								if not header_parsed then
									stderr.put_string ("%NError parsing message header: ")
									stderr.put_string (message_number.out)
									stderr.put_string ("%N")
									stderr.put_string ("==========begin==========%N")
									stderr.put_string (client.response.current_message.header)
									stderr.put_string ("==========end==========%N%N")
								end
							end
							if parser.part.header.has (once_prev_content_type) then
								fix_content_type (parser.part.header, parser.part.header.item (once_prev_content_type).value)
							end
							parser.set_input_buffer (parser.new_string_buffer (client.response.current_message.body))
							parser.reset_parsing_errors
							parser.parse_body
							body_parsed :=
								not parser.syntax_error and then
								parser.parsing_errors = 0
							debug ("emc")
								if not body_parsed then
									stderr.put_string ("%NError parsing message body: ")
									stderr.put_string (message_number.out)
									stderr.put_string ("%N")
									stderr.put_string ("==========begin==========%N")
									if client.response.current_message.body.count <= 1024 then
										stderr.put_string (client.response.current_message.body)
									else
										stderr.put_string (client.response.current_message.body.substring (1, 1024))
										stderr.put_string ("...%N")
									end
									stderr.put_string ("==========end==========%N%N")
								end
							end
							a_tokenizer.start_new_text
							tokenize_header (a_tokenizer, parser.part.header)
							if not header_parsed then
								tokenize_bad_header (a_tokenizer)
							end
							if body_parsed then
								-- Only scan body when we could parse it
								-- correctly.  Else we might end up with all
								-- kinds of encoded tokens, that are no use
								-- anyway.
								tokenize_text_bodies (a_tokenizer, message_number, parser.part)
							else
								tokenize_bad_body (a_tokenizer)
							end
						end
					else
						stderr.put_string ("Skipped body of message ")
						stderr.put_string (message_number.out)
						stderr.put_string (" because size is too large: ")
						stderr.put_string (client.response.current_message.size.out)
						stderr.put_string (" bytes.%N")
					end
				else
					stderr.put_string ("Failure to retrieve header ")
					stderr.put_string (message_number.out)
					stderr.put_string (": ")
					if client.response.response_text /= Void then
						stderr.put_string (client.response.response_text)
					end
					stderr.put_string ("%N")
				end
				message_number := message_number - 1
			end
			fd_stdout.put_string ("Completed 100%%%N")

			client.close_mailbox
		end


feature -- Token info

	dump_tokens (a_file_name: STRING; a_count: INTEGER; hash_table: DS_HASH_TABLE [INTEGER, STRING]) is
			-- Dump tokens to a text file.
		require
			a_file_name_not_empty: a_file_name /= Void and then not a_file_name.is_empty
			positive_count: a_count > 0
			hash_table_not_void: hash_table /= Void
		local
			file: STDC_TEXT_FILE
		do
			create file.create_write (a_file_name)
			file.put_string (a_count.out)
			file.write_character ('%N')
			file.put_string (hash_table.count.out)
			file.write_character ('%N')
			from
				hash_table.start
			until
				hash_table.after
			loop
				file.put_string (hash_table.item_for_iteration.out)
				file.write_character (' ')
				file.put_string (hash_table.key_for_iteration)
				file.write_character ('%N')
				hash_table.forth
			end
			file.close
		end

	print_statistics (a_probability_table: PG_PROBABILITY_TABLE) is
			-- Print statistics about tokens scanned.
		require
			probability_table_not_void: a_probability_table /= Void
		do
			if configuration.minimum_occurrences > 1 then
				print (format ("Dumped only tokens that occur at least $i times.%N", <<integer_cell (configuration.minimum_occurrences)>>))
			end
			print (format ("$i messages scanned for non-spam tokens.%N", <<integer_cell (a_probability_table.nonspam_messages_count)>>))
			print (format ("$i messages scanned for spam tokens.%N", <<integer_cell (a_probability_table.spam_messages_count)>>))
			print (format ("Probability of word not in token list to be spam is $f.%N", <<double_cell (a_probability_table.no_data_points_value)>>))
		end


feature -- Error handling

	assert (msg: STRING; ok: BOOLEAN) is
			-- Stop if not `ok'.
		do
			if not ok then
				exceptions.raise (msg)
			end
		end


feature {NONE} -- Implementation

	fix_content_type (a_header: EPX_MIME_HEADER; original_content_type: STRING) is
			-- SpamAssassin has changed the Content-Type field. In
			-- `original_content_type' is the original content. Fix
			-- `a_header' so its gets the original Content-Type field
			-- again.
		require
			has_content_type: a_header.content_type /= Void
			original_content_type_not_empty: original_content_type /= Void and then not original_content_type.is_empty
		local
			parser: EPX_MIME_PARSER
			dummy_header: STRING
		do
			create parser.make
			dummy_header := field_name_content_type + ": " + original_content_type + "%N%N"
			parser.set_input_buffer (parser.new_string_buffer (dummy_header))
			parser.parse_header
				check
					parsing_ok: not parser.syntax_error
				end
			if
				not parser.syntax_error and then
				parser.parsing_errors = 0 and then
				parser.part.header.content_type /= Void
			then
				a_header.delete_field (field_name_content_type)
				a_header.add_field (parser.part.header.content_type)
			end
		end


feature {NONE} -- Once strings

	once_prev_content_type: STRING is "X-Spam-Prev-Content-Type"


invariant

	configuration_not_void: configuration /= Void

end
